html {
    height: 100%;
}

#grid {
    position: relative;
    background-repeat: no-repeat;
    background-size: cover;
}

#grid .tile,
#grid .object,
#grid .object .shape {
    position:absolute;
}

/* Logo */

#matrx_logo {
    width: 103px;
    height: 43px;
    border-radius: 2px;
    padding: 2px 5px;
}


/* Toggle the start button to pause or vice versa if pressed */
.hidden {
    display: none;
}

.agent_menu_preview,
.god_preview {
    display: inline-block;
    margin-right: 10px;
    width: 30px;
    height: 30px;
}

.agent_menu_preview img {
    max-height: 30px;
    max-width: 30px;
    height: auto;
    width: auto;
}

#agent_dropdown .dropdown-item,
#view_mode .dropdown-item {
    padding: .4rem 1.5rem;
    display: flex;
}


/* Selection styling */
.matrx_object_selected {
    width: 110%;
    z-index: 999;
    position: absolute;
    height: 110%;
    left: -5%;
    top: -5%;
    pointer-events: none;
}



/* Chat styling */
.new_chat_option {
    cursor: pointer;
}

.chat-notification {
    background: #ffb100;
    border-radius: 50%;
    height: 8px;
    width: 8px;
    margin-left: 5px;
    margin-bottom: 1px;
}


.toolbar {
  padding-top: 0.1em;
  padding-bottom: 0.1em;
  padding-left: 1em;
  /* position: sticky; */
  z-index: 100000000000;
}

.row {
  margin-right: 0;
}

.tile.draw_mode:hover {
  cursor: url("pencil-alt.svg") 0 15, pointer;
  background-color: indianred;
}

.tile.erase_mode:hover {
  cursor: url("eraser.svg") 0 15, pointer;
}

.agent {
  z-index: 1;
  color: #fff;
  background-color: #17a2b8;
}

.agent:hover {
  cursor: pointer;
  background-color: #138496;
}

.grid {
  background-color: lightgray;
}

.contact {
  background-color: #f8f9fa;
  color: #212529;
  text-align: center;
  vertical-align: middle;
  border: 1px solid transparent;
  padding: 0.375rem 0.75rem;
  border-radius: 0.25rem;
  transition: color 0.15s ease-in-out, background-color 0.15s ease-in-out, border-color 0.15s ease-in-out, box-shadow 0.15s ease-in-out;
  margin-bottom: 0.5em;
}

.contact:hover {
  cursor: pointer;
  background-color: #e2e6ea;
}

.contact_active {
  background-color: #e2e6ea;
}

#messages {
    overflow-y: auto;
    max-height: 50vh;
}

.message_you {
    color: #fff;
    position: relative;
    background: #00aabb;
    border-radius: .4em;
    padding: 0.375rem 0.75rem;
    margin-bottom: 0.5em;
    width: fit-content;
    margin-left: auto;
    margin-right: 10px;
}

.message_you:after {
	content: '';
	position: absolute;
	right: 0;
	top: 50%;
	width: 0;
	height: 0;
	border: 8px solid transparent;
	border-left-color: #00aabb;
	border-right: 0;
	border-top: 0;
	margin-top: -4px;
	margin-right: -8px;
}

.message_other {
	position: relative;
	background: #e2e6ea;
	border-radius: .4em;
    padding: 0.375rem 0.75rem;
    margin-bottom: 0.5em;
    width: fit-content;
    margin-left: 10px;
}

.message_other:after {
  content: '';
	position: absolute;
	left: 0;
	top: 50%;
	width: 0;
	height: 0;
	border: 8px solid transparent;
	border-right-color: #e2e6ea;
	border-left: 0;
	border-top: 0;
	margin-top: -4px;
	margin-left: -8px;
}

.form-control {
  margin-bottom: 0.5em;
}

h5 {
  margin-top: 0.5em;
  margin-bottom: 0.5em;
}

#chat {
  background-color: white;
  padding-bottom: 1em;
  padding-left: 0.5em;
  padding-right: 1em;
  z-index: 100000000000;
}

.object:hover {
    cursor: pointer;
}

.chat-mssg-sender {
    font-weight: bold;
}

.matrx_object_busy {
    position: absolute;
}