body {
    overflow: hidden;
}

#container {
    height: 100%;
    width: 100%;
    position: absolute;
}

.area {
    border: 1px dashed black;
    height: 80%;
    width: 80%;
    position: absolute;
}

.circle {
    height: 200px;
    width: 200px;
    background-color: navajowhite;
    border-radius: 50%;
    display: inline-block;
    position: relative;
}

.rectangle {
    background-color: navajowhite;
    height: 150px;
    width: 250px;
    position: relative;
}

.text-area {
    float: right;
    position: relative;
}
