#  __  __       _______ _____  __   __
# |  \/  |   /\|__   __|  __ \ \ \ / /
# | \  / |  /  \  | |  | |__) | \ V / 
# | |\/| | / /\ \ | |  |  _  /   > <  
# | |  | |/ ____ \| |  | | \ \  / . \ 
# |_|  |_/_/    \_\_|  |_|  \_\/_/ \_\
#

__title__ = 'matrx'
__description__ = "A Python package for the rapid development and " \
                  "evaluation of human-agent teaming concepts."
__url__ = 'https://matrx-software.com'
__doc_url__ = 'http://docs.matrx-software.com/en/latest/'
__source_url__ = 'https://github.com/matrx-software/matrx'
__version__ = '1.1.1'
__author__ = 'MATRX Team at TNO.nl'
__author_email__ = 'info@matrx.com'
__license__ = 'MIT License'
__copyright__ = 'Copyright 2020 TNO'
__cake__ = u'\u2728 \U0001f370 \u2728'
